@@ -14,40 +14,91 @@ import socket
 
 from . import keys
 from .compat import text_type
+from .exceptions import GNTPError
 
 
-__version__ = '0.4'
-__all__ = ['notify', 'Notifier', 'RawIcon', 'SocketCallback', 'Event']
+__version__ = '0.5'
+__all__ = ['notify', 'publish', 'subscribe', 'Event', 'Notifier', 'Publisher',
+           'RawIcon', 'Resource', 'SocketCallback', 'Subscriber']
 
 
 SUPPORTED_VERSIONS = ['1.0']
+DEFAULT_PORT = 23053
+DEFAULT_TIMEOUT = 10
+DEFAULT_TTL = 60
 MAX_MESSAGE_SIZE = 4096
 MAX_LINE_SIZE = 1024
 LINE_DELIMITER = b'\r\n'
 SECTION_DELIMITER = SECTION_BODY_START = SECTION_BODY_END = b'\r\n'
 MESSAGE_DELIMITER = b'\r\n\r\n'
+MESSAGE_DELIMITER_SIZE = len(MESSAGE_DELIMITER)
 RESPONSE_INFORMATION_LINE_RE = re.compile(
     b'GNTP/([^ ]+) (-OK|-ERROR|-CALLBACK) NONE')
+CUSTOM_HEADER_PREFIX = 'X-'
+APP_SPECIFIC_HEADER_PREFIX = 'Data-'
 
 
-def notify(app_name, event_name, title, text=''):
-    """Register a notifier and send a notification at a time.
+def publish(app_name, event_name, title, text=''):
+    """Register a publisher and send a notification at a time.
 
     :param app_name: the name of the application.
     :param event_name: the name of the notification.
     :param title: the title of the notification.
     :param text: the text of the notification.  Defaults to ``''``.
     """
-    notifier = Notifier(app_name, coerce_to_events([event_name]))
-    notifier.register()
-    notifier.notify(event_name, title, text)
+    publisher = Publisher(app_name, coerce_to_events([event_name]))
+    publisher.register()
+    publisher.publish(event_name, title, text)
 
 
-class Notifier(object):
-    """Notifier of Growl Notification Transport Protocol (GNTP).
+def notify(app_name, event_name, title, text=''):
+    """Deprecated notify function."""
+    import warnings
+    warnings.warn('notify function is deprecated, use publish function'
+                  ' instead', DeprecationWarning, stacklevel=2)
+    publish(app_name, event_name, title, text)
 
-    Currently this class supports ``REGISTER`` and ``NOTIFY`` requests.  They
-    are sent by :meth:`register()` and :meth:`notify()` methods respectively.
+
+def subscribe(id_, name, hub, password, port=DEFAULT_PORT):
+    """Send a subscription request and return ttl.
+
+    :param id_: the unique id of the subscriber.
+    :param name: the name of the subscriber.
+    :param hub: the subscribed-to machine.  If a string is given, it is used as
+                a host of the hub and default port number `23053` is used.
+                If host-port tuple is given, it is used directly.
+    :param password: the password of the subscribed-to machine.
+    :param port: the port number of the subscriber.  Defaults to `23053`.
+    """
+    subscriber = Subscriber(id_, name, hub, password, port=port)
+    subscriber.subscribe()
+    return subscriber.ttl
+
+
+class BaseApp(object):
+    """Base class for applications.
+
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
+    :param gntp_client_class: GNTP client class.  If it is `None`,
+                              :class:`GNTPClient` is used.  Defaults to `None`.
+    """
+
+    def __init__(self, custom_headers=None, app_specific_headers=None,
+                 gntp_client_class=None, **kwargs):
+        self.custom_headers = custom_headers or []
+        self.app_specific_headers = app_specific_headers or []
+        if gntp_client_class is None:
+            gntp_client_class = GNTPClient
+        self.gntp_client = gntp_client_class(**kwargs)
+
+
+class Publisher(BaseApp):
+    """Publisher of Growl Notification Transport Protocol (GNTP).
+
+    This class supports ``REGISTER`` and ``NOTIFY`` requests.  They
+    are sent by :meth:`register()` and :meth:`publish()` methods respectively.
     These methods can accept the optional final callback as `callback` keyword
     argument, which run after closing the connection with the GNTP server.
 
@@ -64,37 +115,41 @@ class Notifier(object):
 
     :param name: the name of the application.
     :param event_defs: the definitions of the notifications.
-    :param icon: url string or an instance of subclass of :class:`BaseIcon`
-                 for the icon of the application.  Defaults to `None`.
+    :param icon: url string or an instance of :class:`Resource` for the icon of
+                 the application.  Defaults to `None`.
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
     :param gntp_client_class: GNTP client class.  If it is `None`,
                               :class:`GNTPClient` is used.  Defaults to `None`.
 
     .. note:: In Growl 1.3.3, `icon` of url string does not work.
     """
 
-    def __init__(self, name, event_defs, icon=None, gntp_client_class=None,
-                 **kwargs):
+    def __init__(self, name, event_defs, icon=None, custom_headers=None,
+                 app_specific_headers=None, gntp_client_class=None, **kwargs):
         self.name = name
         self.icon = icon
         self.events = coerce_to_events(event_defs)
         if not self.events:
             raise GNTPError('You have to set at least one notification type')
-        if gntp_client_class is None:
-            gntp_client_class = GNTPClient
-        self.gntp_client = gntp_client_class(**kwargs)
+        BaseApp.__init__(self, custom_headers, app_specific_headers,
+                         gntp_client_class, **kwargs)
 
     def register(self, callback=None):
-        """Register this notifier to the GNTP server.
+        """Register this publisher to the GNTP server.
 
         :param callback: the callback run after closing the connection with
                          the GNTP server.  Defaults to `None`.
         """
-        request = RegisterRequest(self.name, self.icon, self.events)
+        request = RegisterRequest(self.name, self.icon, self.events,
+                                  self.custom_headers,
+                                  self.app_specific_headers)
         self.gntp_client.process_request(request, callback)
 
-    def notify(self, name, title, text='', id_=None, sticky=False,
-               priority=0, icon=None, coalescing_id=None, callback=None,
-               gntp_callback=None, **socket_callback_options):
+    def publish(self, name, title, text='', id_=None, sticky=False,
+                priority=0, icon=None, coalescing_id=None, callback=None,
+                gntp_callback=None, **socket_callback_options):
         """Send a notification to the GNTP server.
 
         :param name: the name of the notification.
@@ -107,8 +162,8 @@ class Notifier(object):
         :param priority: the display hint for the receiver which may be
                          ignored.  A higher number indicates a higher priority.
                          Valid values are between -2 and 2, defaults to `0`.
-        :param icon: url string or an instance of subclass of :class:`BaseIcon`
-                     to display with the notification.  Defaults to `None`.
+        :param icon: url string or an instance of :class:`Resource` to display
+                     with the notification.  Defaults to `None`.
         :param coalescing_id: if set, should contain the value of the `id_` of
                               a previously-sent notification.
                               This serves as a hint to the notification system
@@ -137,13 +192,90 @@ class Notifier(object):
                                     coalescing_id=coalescing_id,
                                     gntp_callback=gntp_callback,
                                     **socket_callback_options)
-        request = NotifyRequest(self.name, notification)
+        request = NotifyRequest(self.name, notification,
+                                self.custom_headers,
+                                self.app_specific_headers)
         self.gntp_client.process_request(
             request, callback, socket_callback=notification.socket_callback)
 
 
-class GNTPError(Exception):
-    """Base class for GNTP exception."""
+class Notifier(Publisher):
+    """Deprecated Notifier of Growl Notification Transport Protocol (GNTP)."""
+
+    def __init__(self, name, event_defs, icon=None, custom_headers=None,
+                 app_specific_headers=None, gntp_client_class=None, **kwargs):
+        import warnings
+        warnings.warn('Notifier is deprecated, use Publisher instead',
+                      DeprecationWarning, stacklevel=2)
+        Publisher.__init__(self, name, event_defs, icon,
+                           custom_headers, app_specific_headers,
+                           gntp_client_class, **kwargs)
+
+    def notify(self, name, title, text='', id_=None, sticky=False,
+               priority=0, icon=None, coalescing_id=None, callback=None,
+               gntp_callback=None, **socket_callback_options):
+        """Send a notification to the GNTP server."""
+        import warnings
+        warnings.warn('notify method is deprecated, use publish method'
+                      ' instead', DeprecationWarning, stacklevel=2)
+        self.publish(name, title, text, id_, sticky, priority, icon,
+                     coalescing_id, callback, gntp_callback,
+                     **socket_callback_options)
+
+
+class Subscriber(BaseApp):
+    """Subscriber of Growl Notification Transport Protocol (GNTP).
+
+    This class supports ``SUBSCRIBE`` request.
+
+    :param id_: the unique id of the subscriber.
+    :param name: the name of the subscriber.
+    :param hub: the subscribed-to machine.  If a string is given, it is used as
+                a host of the hub and default port number `23053` is used.
+                If host-port tuple is given, it is used directly.
+    :param password: the password of the subscribed-to machine.
+    :param port: the port number of the subscriber.  Defaults to `23053`.
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
+    :param gntp_client_class: GNTP client class.  If it is `None`,
+                              :class:`GNTPClient` is used.  Defaults to `None`.
+    """
+
+    def __init__(self, id_, name, hub, password, port=DEFAULT_PORT,
+                 custom_headers=None, app_specific_headers=None,
+                 gntp_client_class=None, **kwargs):
+        self.id_ = id_
+        self.name = name
+        if isinstance(hub, (bytes, text_type)):
+            self.hub = (hub, DEFAULT_PORT)
+        else:
+            self.hub = hub
+        self.password = password
+        self.port = port
+        self.ttl = DEFAULT_TTL
+        BaseApp.__init__(self, custom_headers, app_specific_headers,
+                         gntp_client_class, host=self.hub[0], port=self.hub[1],
+                         password=self.password, **kwargs)
+
+    def subscribe(self, callback=None):
+        """Send a subscription request.
+
+        If `callback` is `None`, :meth:`store_ttl` is used and :attr:`ttl` is
+        updated by ``Subscription-TTL`` value of the response.
+
+        :param callback: the callback run after closing the connection with
+                         the GNTP server.  Defaults to `None`.
+        """
+        request = SubscribeRequest(self.id_, self.name, self.port,
+                                   self.custom_headers,
+                                   self.app_specific_headers)
+        self.gntp_client.process_request(request, callback or self.store_ttl)
+
+    def store_ttl(self, response):
+        """Update :attr:`ttl` attribute."""
+        ttl = response.headers['Subscription-TTL']
+        self.ttl = int(ttl)
 
 
 class Event(object):
@@ -155,9 +287,9 @@ class Event(object):
                          Preferences.  Defaults to `None`.
     :param enabled: indicates if the notification should be enabled by
                     default.  Defaults to `True`.
-    :param icon: url string or an instance of subclass of :class:`BaseIcon`
-                 for the default icon to display with the notifications of this
-                 notification type.  Defaults to `None`.
+    :param icon: url string or an instance of :class:`Resource` for the default
+                 icon to display with the notifications of this notification
+                 type.  Defaults to `None`.
 
     .. note:: In Growl 1.3.3, `icon` does not work.
     """
@@ -166,7 +298,7 @@ class Event(object):
         self.name = name
         self.display_name = display_name
         self.enabled = enabled
-        self.icon = coerce_to_icon(icon)
+        self.icon = icon
 
 
 class Notification(object):
@@ -181,7 +313,7 @@ class Notification(object):
         self.id_ = id_
         self.sticky = sticky
         self.priority = priority
-        self.icon = coerce_to_icon(icon)
+        self.icon = icon
         self.coalescing_id = coalescing_id
         self.callback = coerce_to_callback(gntp_callback,
                                            **socket_callback_options)
@@ -192,43 +324,99 @@ class Notification(object):
             return self.callback
 
 
-class RegisterRequest(object):
+class BaseRequest(object):
+    """Abstract base class for GNTP request.
+
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
+    """
+    #: Request message type.  Subclasses must override this attribute.
+    message_type = None
+
+    def __init__(self, custom_headers=None, app_specific_headers=None):
+        self.custom_headers = custom_headers or []
+        self.app_specific_headers = app_specific_headers or []
+
+    def write_into(self, writer):
+        """Subclasses must call this method first to serialize their
+        message."""
+        writer.write_base_request(self)
+
+
+class RegisterRequest(BaseRequest):
     """Represent ``REGISTER`` request.
 
     :param app_name: the name of the application.
-    :param app_icon: url string or an instance of subclass of :class:`BaseIcon`
-                     for the icon of the application.
+    :param app_icon: url string or an instance of :class:`Resource` for the
+                     icon of the application.
     :param events: the list of :class:`Event` instances.
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
 
     .. note:: In Growl 1.3.3, `app_icon` of url string does not work.
     """
     message_type = 'REGISTER'
 
-    def __init__(self, app_name, app_icon, events):
+    def __init__(self, app_name, app_icon, events, custom_headers=None,
+                 app_specific_headers=None):
+        BaseRequest.__init__(self, custom_headers, app_specific_headers)
         self.app_name = app_name
-        self.app_icon = coerce_to_icon(app_icon)
+        self.app_icon = app_icon
         self.events = events
 
     def write_into(self, writer):
+        BaseRequest.write_into(self, writer)
         writer.write_register_request(self)
 
 
-class NotifyRequest(object):
+class NotifyRequest(BaseRequest):
     """Represent ``NOTIFY`` request.
 
     :param app_name: the name of the application.
     :param notification: :class:`Notification` instance.
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
     """
     message_type = 'NOTIFY'
 
-    def __init__(self, app_name, notification):
+    def __init__(self, app_name, notification, custom_headers=None,
+                 app_specific_headers=None):
+        BaseRequest.__init__(self, custom_headers, app_specific_headers)
         self.app_name = app_name
         self.notification = notification
 
     def write_into(self, writer):
+        BaseRequest.write_into(self, writer)
         writer.write_notify_request(self)
 
 
+class SubscribeRequest(BaseRequest):
+    """Represent ``SUBSCRIBE`` request.
+
+    :param id_: the unique id of the subscriber.
+    :param name: the name of the subscriber.
+    :param port: the port number of the subscriber.
+    :param custom_headers: the list of key-value tuples for Custom Headers.
+    :param app_specific_headers: the list of key-value tuples for App-Specific
+                                 Headers.
+    """
+    message_type = 'SUBSCRIBE'
+
+    def __init__(self, id_, name, port, custom_headers=None,
+                 app_specific_headers=None):
+        BaseRequest.__init__(self, custom_headers, app_specific_headers)
+        self.id_ = id_
+        self.name = name
+        self.port = port
+
+    def write_into(self, writer):
+        BaseRequest.write_into(self, writer)
+        writer.write_subscribe_request(self)
+
+
 class Response(object):
     """Base class for GNTP response.
 
@@ -334,8 +522,9 @@ class GNTPClient(object):
                              `None`.
     """
 
-    def __init__(self, host='localhost', port=23053, timeout=10,
-                 password=None, key_hashing=keys.SHA256, encryption=None,
+    def __init__(self, host='localhost', port=DEFAULT_PORT,
+                 timeout=DEFAULT_TIMEOUT, password=None,
+                 key_hashing=keys.SHA256, encryption=None,
                  connection_class=None):
         self.address = (host, port)
         self.timeout = timeout
@@ -346,7 +535,7 @@ class GNTPClient(object):
                             ' least encryption key size (%s:%d)' % (
                     key_hashing.algorithm_id, key_hashing.key_size,
                     encryption.algorithm_id, encryption.key_size))
-        self.packer_factory = RequestPackerFactory(password, key_hashing,
+        self.packer_factory = MessagePackerFactory(password, key_hashing,
                                                    encryption)
 
     def process_request(self, request, callback, **kwargs):
@@ -375,7 +564,7 @@ def generate_messages(sock, size=1024):
             break
         pos = buf.find(MESSAGE_DELIMITER)
         if ((pos < 0 and len(buf) >= MAX_MESSAGE_SIZE) or
-            (pos > MAX_MESSAGE_SIZE - 4)):
+            (pos > MAX_MESSAGE_SIZE - MESSAGE_DELIMITER_SIZE)):
             raise GNTPError('too large message: %r' % buf)
         elif pos > 0:
             pos += 4
@@ -433,65 +622,37 @@ def coerce_to_events(items):
     return results
 
 
-class BaseIcon(object):
-    """Abstract base class for icon."""
+class Resource(object):
+    """Class for <uniqueid> data types.
 
-    def __init__(self, data):
-        self.data = data
-
-    def resource(self):
-        """Used to write header of this instance into the writer."""
-        raise NotImplementedError
-
-    def write_into(self, writer):
-        """Used to write section of this instance into the writer."""
-        raise NotImplementedError
-
-
-class RawIcon(BaseIcon):
-    """Class for icon of <uniqueid> data type.
-
-    :param data: bytes of icon image.
+    :param data: the binary content.
     """
 
     def __init__(self, data):
-        BaseIcon.__init__(self, data)
-        self._unique_id = None
+        self.data = data
+        self._unique_value = None
+
+    def unique_value(self):
+        """Return the <uniquevalue> value."""
+        if self.data is not None and self._unique_value is None:
+            self._unique_value = \
+                hashlib.md5(self.data).hexdigest().encode('utf-8')
+        return self._unique_value
 
     def unique_id(self):
-        """Return the value of <uniqueid> data type."""
-        if self.data is not None and self._unique_id is None:
-            self._unique_id = hashlib.md5(self.data).hexdigest()
-        return self._unique_id
-
-    def resource(self):
-        """Return the resource string used in request headers."""
+        """Return the <uniqueid> value."""
         if self.data is not None:
-            return 'x-growl-resource://%s' % self.unique_id()
-
-    def write_into(self, writer):
-        if self.data is not None:
-            writer.write_raw_icon(self)
+            return b'x-growl-resource://' + self.unique_value()
 
 
-class URLIcon(BaseIcon):
-    """Class for icon of <url> data type."""
+class RawIcon(Resource):
+    """Deprecated icon class."""
 
-    def __init__(self, url):
-        BaseIcon.__init__(self, url)
-
-    def resource(self):
-        return self.data
-
-    def write_into(self, writer):
-        pass
-
-
-def coerce_to_icon(obj):
-    if obj is None or isinstance(obj, BaseIcon):
-        return obj
-    if isinstance(obj, (bytes, text_type)):
-        return URLIcon(obj)
+    def __init__(self, data):
+        import warnings
+        warnings.warn('RawIcon is deprecated, use Resource instead',
+                      DeprecationWarning, stacklevel=2)
+        Resource.__init__(self, data)
 
 
 class SocketCallback(object):
@@ -602,8 +763,8 @@ class _NullCipher(object):
 NullCipher = _NullCipher()
 
 
-class RequestPackerFactory(object):
-    """The factory of :class:`RequestPacker`.
+class MessagePackerFactory(object):
+    """The factory of :class:`MessagePacker`.
 
     If `password` is None, `hashing` and `encryption` are ignored.
     """
@@ -614,14 +775,14 @@ class RequestPackerFactory(object):
         self.encryption = (password and encryption) or NullCipher
 
     def create(self):
-        """Create an instance of :class:`RequestPacker` and return it."""
+        """Create an instance of :class:`MessagePacker` and return it."""
         key = self.password and self.hashing.key(self.password)
         cipher = self.encryption and self.encryption.cipher(key)
-        return RequestPacker(key, cipher)
+        return MessagePacker(key, cipher)
 
 
-class RequestPacker(object):
-    """The serializer of requests.
+class MessagePacker(object):
+    """The serializer for messages.
 
     `key` and `cipher` have random-generated salt and iv respectively.
 
@@ -635,13 +796,20 @@ class RequestPacker(object):
 
     def pack(self, request):
         """Return utf-8 encoded request message."""
-        return (self._information_line(request) +
+        return (InformationLinePacker(self.key, self.cipher).pack(request) +
                 LINE_DELIMITER +
-                self._headers(request) +
-                self._sections(request) +
+                HeaderPacker(self.cipher).pack(request) +
+                SectionPacker(self.cipher).pack(request) +
                 LINE_DELIMITER)
 
-    def _information_line(self, request):
+
+class InformationLinePacker(object):
+
+    def __init__(self, key, cipher):
+        self.key = key
+        self.cipher = cipher
+
+    def pack(self, request):
         """Return utf-8 encoded information line."""
         result = (b'GNTP/1.0 ' +
                   request.message_type.encode('utf-8') +
@@ -658,57 +826,62 @@ class RequestPacker(object):
                        self.key.salt_hex)
         return result
 
-    def _headers(self, request):
+
+class HeaderPacker(object):
+
+    def __init__(self, cipher):
+        self.writer = io.BytesIO()
+        self.cipher = cipher
+
+    def pack(self, request):
         """Return utf-8 encoded headers."""
-        writer = io.BytesIO()
-        header_writer = HeaderWriter(writer)
-        request.write_into(header_writer)
-        headers = writer.getvalue()
+        request.write_into(self)
+        headers = self.writer.getvalue()
         result = self.cipher.encrypt(headers)
         if self.cipher.algorithm is not None:
             result += LINE_DELIMITER
         return result
 
-    def _sections(self, request):
-        """Return utf-8 encoded message body."""
-        writer = io.BytesIO()
-        header_writer = HeaderWriter(writer)
-        section_writer = SectionWriter(writer, header_writer, self.cipher)
-        request.write_into(section_writer)
-        return writer.getvalue()
+    def write_base_request(self, request):
+        self._write_additional_headers(request.custom_headers,
+                                       CUSTOM_HEADER_PREFIX)
+        self._write_additional_headers(request.app_specific_headers,
+                                       APP_SPECIFIC_HEADER_PREFIX)
 
-
-class HeaderWriter(object):
-
-    def __init__(self, writer):
-        self.writer = writer
+    def _write_additional_headers(self, headers, prefix):
+        for key, value in headers:
+            if not key.startswith(prefix):
+                key = prefix + key
+            self.write(key.encode('utf-8'), value)
 
     def write_register_request(self, request):
         self.write(b'Application-Name', request.app_name)
-        self.write_icon(request.app_icon, b'Application-Icon')
+        self.write(b'Application-Icon', request.app_icon)
         self.write(b'Notifications-Count', len(request.events))
         for event in request.events:
             self.writer.write(LINE_DELIMITER)
-            self.write_event(event)
+            self.write(b'Notification-Name', event.name)
+            self.write(b'Notification-Display-Name', event.display_name)
+            self.write(b'Notification-Enabled', event.enabled)
+            self.write(b'Notification-Icon', event.icon)
 
     def write_notify_request(self, request):
         self.write(b'Application-Name', request.app_name)
-        self.write_notification(request.notification)
+        self._write_notification(request.notification)
 
-    def write_event(self, event):
-        self.write(b'Notification-Name', event.name)
-        self.write(b'Notification-Display-Name', event.display_name)
-        self.write(b'Notification-Enabled', event.enabled)
-        self.write_icon(event.icon, b'Notification-Icon')
+    def write_subscribe_request(self, request):
+        self.write(b'Subscriber-ID', request.id_)
+        self.write(b'Subscriber-Name', request.name)
+        self.write(b'Subscriber-Port', request.port)
 
-    def write_notification(self, notification):
+    def _write_notification(self, notification):
         self.write(b'Notification-Name', notification.name)
         self.write(b'Notification-ID', notification.id_)
         self.write(b'Notification-Title', notification.title)
         self.write(b'Notification-Text', notification.text)
         self.write(b'Notification-Sticky', notification.sticky)
         self.write(b'Notification-Priority', notification.priority)
-        self.write_icon(notification.icon, b'Notification-Icon')
+        self.write(b'Notification-Icon', notification.icon)
         self.write(b'Notification-Coalescing-ID', notification.coalescing_id)
         if notification.callback is not None:
             notification.callback.write_into(self)
@@ -721,16 +894,14 @@ class HeaderWriter(object):
     def write_url_callback(self, callback):
         self.write(b'Notification-Callback-Target', callback.url)
 
-    def write_icon(self, icon, name):
-        if icon is not None:
-            self.write(name, icon.resource())
-
     def write(self, name, value):
-        """Write utf-8 encoded headers into writer.
+        """Write utf-8 encoded header into writer.
 
         :param name: the name of the header.
         :param value: the value of the header.
         """
+        if isinstance(value, Resource):
+            value = value.unique_id()
         if value is not None:
             if not isinstance(value, bytes):
                 value = text_type(value).encode('utf-8')
@@ -740,46 +911,49 @@ class HeaderWriter(object):
             self.writer.write(LINE_DELIMITER)
 
 
-class SectionWriter(object):
+class SectionPacker(object):
 
-    def __init__(self, writer, header_writer, cipher):
-        self.writer = writer
-        self.header_writer = header_writer
+    def __init__(self, cipher):
+        self.writer = io.BytesIO()
         self.cipher = cipher
 
+    def pack(self, request):
+        """Return utf-8 encoded message body."""
+        request.write_into(self)
+        return self.writer.getvalue()
+
+    def write_base_request(self, request):
+        for _, value in request.custom_headers:
+            self.write(value)
+        for _, value in request.app_specific_headers:
+            self.write(value)
+
     def write_register_request(self, request):
-        self.write_icon(request.app_icon)
+        self.write(request.app_icon)
         for event in request.events:
-            self.write_event(event)
+            self.write(event.icon)
 
     def write_notify_request(self, request):
-        self.write_notification(request.notification)
+        self.write(request.notification.icon)
 
-    def write_event(self, event):
-        self.write_icon(event.icon)
+    def write_subscribe_request(self, request):
+        pass
 
-    def write_notification(self, notification):
-        self.write_icon(notification.icon)
-
-    def write_raw_icon(self, icon):
-        icon_data = self.cipher.encrypt(icon.data)
-        headers = [(b'Identifier', icon.unique_id()),
-                   (b'Length', len(icon_data))]
-        self.write(headers, icon_data)
-
-    def write_icon(self, icon):
-        if icon is not None:
-            icon.write_into(self)
-
-    def write(self, headers, body):
-        """Write utf-8 encoded sections into writer.
+    def write(self, resource):
+        """Write utf-8 encoded resource into writer.
 
         :param headers: the iterable of (`name`, `value`) tuple of the header.
         :param body: bytes of section body.
         """
-        self.writer.write(SECTION_DELIMITER)
-        for name, value in headers:
-            self.header_writer.write(name, value)
-        self.writer.write(SECTION_BODY_START)
-        self.writer.write(body)
-        self.writer.write(SECTION_BODY_END)
+        if isinstance(resource, Resource) and resource.data is not None:
+            data = self.cipher.encrypt(resource.data)
+            self.writer.write(SECTION_DELIMITER)
+            self.writer.write(b'Identifier: ')
+            self.writer.write(resource.unique_value())
+            self.writer.write(LINE_DELIMITER)
+            self.writer.write(b'Length: ')
+            self.writer.write(text_type(len(data)).encode('utf-8'))
+            self.writer.write(LINE_DELIMITER)
+            self.writer.write(SECTION_BODY_START)
+            self.writer.write(data)
+            self.writer.write(SECTION_BODY_END)
